local speed_mutliplier = 2

---@type ObjectHitbox
local sGoombaHitbox = {
    interactType      = INTERACT_BOUNCE_TOP,
    downOffset        = 0,
    damageOrCoinValue = 1,
    health            = 0,
    numLootCoins      = 1,
    radius            = 72,
    height            = 50,
    hurtboxRadius     = 42,
    hurtboxHeight     = 40
}

local sGoombaAttackHandlers = {
    -- regular and tiny
    {
        --[[ATTACK_PUNCH                 = ]] ATTACK_HANDLER_KNOCKBACK,
        --[[ATTACK_KICK_OR_TRIP          = ]] ATTACK_HANDLER_KNOCKBACK,
        --[[ATTACK_FROM_ABOVE            = ]] ATTACK_HANDLER_SQUISHED,
        --[[ATTACK_GROUND_POUND_OR_TWIRL = ]] ATTACK_HANDLER_SQUISHED,
        --[[ATTACK_FAST_ATTACK           = ]] ATTACK_HANDLER_KNOCKBACK,
        --[[ATTACK_FROM_BELOW            = ]] ATTACK_HANDLER_KNOCKBACK,
    },
    -- huge
    {
        --[[ATTACK_PUNCH                 = ]] ATTACK_HANDLER_SPECIAL_HUGE_GOOMBA_WEAKLY_ATTACKED,
        --[[ATTACK_KICK_OR_TRIP          = ]] ATTACK_HANDLER_SPECIAL_HUGE_GOOMBA_WEAKLY_ATTACKED,
        --[[ATTACK_FROM_ABOVE            = ]] ATTACK_HANDLER_SQUISHED,
        --[[ATTACK_GROUND_POUND_OR_TWIRL = ]] ATTACK_HANDLER_SQUISHED_WITH_BLUE_COIN,
        --[[ATTACK_FAST_ATTACK           = ]] ATTACK_HANDLER_SPECIAL_HUGE_GOOMBA_WEAKLY_ATTACKED,
        --[[ATTACK_FROM_BELOW            = ]] ATTACK_HANDLER_SPECIAL_HUGE_GOOMBA_WEAKLY_ATTACKED,
    },
}

local sGoombaProperties = {
    { scale = 1.5, deathSound = SOUND_OBJ_ENEMY_DEATH_HIGH, drawDistance = 4000, damage = 2 }, -- Normal
    { scale = 3.5, deathSound = SOUND_OBJ_ENEMY_DEATH_LOW, drawDistance = 4000, damage = 2 }, -- Big
    { scale = 0.5, deathSound = SOUND_OBJ_ENEMY_DEATH_HIGH, drawDistance = 1500, damage = 1 }, -- Small
};

---@param param any
---@param case_table table
local function switch(param, case_table)
    local case = case_table[param]
    return case and case() or nil
end

---@param obj Object
---@param hitbox ObjectHitbox
local function obj_set_hitbox(obj, hitbox)
    if not obj or not hitbox then return end
    if obj.oFlags & OBJ_FLAG_30 == 0 then
        obj.oFlags = obj.oFlags | OBJ_FLAG_30

        obj.oInteractType = hitbox.interactType
        obj.oDamageOrCoinValue = hitbox.damageOrCoinValue
        obj.oHealth = hitbox.health
        obj.oNumLootCoins = hitbox.numLootCoins

        cur_obj_become_tangible()
    end

    obj.hitboxRadius = obj.header.gfx.scale.x * hitbox.radius
    obj.hitboxHeight = obj.header.gfx.scale.y * hitbox.height
    obj.hurtboxRadius = obj.header.gfx.scale.x * hitbox.hurtboxRadius
    obj.hurtboxHeight = obj.header.gfx.scale.y * hitbox.hurtboxHeight
    obj.hitboxDownOffset = obj.header.gfx.scale.y * hitbox.downOffset
end

---@param obj Object
---@param hitbox ObjectHitbox
---@param attackedMarioAction integer
---@param attackHandlers integer
---@return boolean
local function obj_handle_attacks(obj, hitbox, attackedMarioAction, attackHandlers)
    if not obj then return false end
    local attackType = 0

    obj_set_hitbox(obj, hitbox)

    if obj_die_if_above_lava_and_health_non_positive() ~= 0 then
        return true
    elseif obj.oInteractStatus & INT_STATUS_INTERACTED ~= 0 then
        if obj.oInteractStatus & INT_STATUS_ATTACKED_MARIO ~= 0 then
            if obj.oAction ~= attackedMarioAction then
                obj.oAction = attackedMarioAction
                obj.oTimer = 0
            end
        else
            attackType = obj.oInteractStatus & INT_STATUS_ATTACK_MASK
            switch (attackHandlers[attackType], {
                --[[[ATTACK_HANDLER_NOP] = function ()
                    -- do nothing
                end,
                [ATTACK_HANDLER_DIE_IF_HEALTH_NON_POSITIVE] = function ()
                    obj_die_if_health_non_positive()
                end,]]
                [ATTACK_HANDLER_KNOCKBACK] = function ()
                    obj_set_knockback_action(attackType)
                end,
                [ATTACK_HANDLER_SQUISHED] = function ()
                    obj_set_squished_action()
                end,
                --[[[ATTACK_HANDLER_SPECIAL_KOOPA_LOSE_SHELL] = function ()
                    shelled_koopa_attack_handler(attackType)
                end,
                [ATTACK_HANDLER_SET_SPEED_TO_ZERO] = function ()
                    obj_set_speed_to_zero()
                end,
                [ATTACK_HANDLER_SPECIAL_WIGGLER_JUMPED_ON] = function ()
                    wiggler_jumped_on_attack_handler()
                end,]]
                [ATTACK_HANDLER_SPECIAL_HUGE_GOOMBA_WEAKLY_ATTACKED] = function ()
                    --huge_goomba_weakly_attacked()
                    obj.oAction = GOOMBA_ACT_ATTACKED_MARIO
                end,
                [ATTACK_HANDLER_SQUISHED_WITH_BLUE_COIN] = function ()
                    obj.oNumLootCoins = -1
                    obj_set_squished_action()
                end,
            })

            obj.oInteractStatus = 0
            return true
        end
    end

    obj.oInteractStatus = 0
    return false
end

---@param obj Object
---@param threshold number
---@param distanceToPlayer number?
---@param angleToPlayer number?
---@return number?, number?
local function treat_far_home_as_mario(obj, threshold, distanceToPlayer, angleToPlayer)
    if not obj then return nil, nil end
    local dx = obj.oHomeX - obj.oPosX
    local dy = obj.oHomeY - obj.oPosY
    local dz = obj.oHomeZ - obj.oPosZ
    local distance = math.sqrt(dx * dx + dy * dy + dz * dz)

    if distance > threshold then
        obj.oAngleToMario = atan2s(dz, dx)
        obj.oDistanceToMario = 25000.0
        if distanceToPlayer then
            distanceToPlayer = obj.oDistanceToMario
        end
        if angleToPlayer then
            angleToPlayer = obj.oAngleToMario
        end
    else
        local player = nearest_player_to_object(obj)
        if player then
            dx = obj.oHomeX - player.oPosX
            dy = obj.oHomeY - player.oPosY
            dz = obj.oHomeZ - player.oPosZ
            distance = math.sqrt(dx * dx + dy * dy + dz * dz)

            if distance > threshold then
                obj.oDistanceToMario = 20000.0
                if distanceToPlayer then
                    distanceToPlayer = obj.oDistanceToMario
                end
            end
        else
            obj.oDistanceToMario = 20000.0
            if distanceToPlayer then
                distanceToPlayer = obj.oDistanceToMario
            end
        end
    end

    return distanceToPlayer, angleToPlayer
end


---@param obj Object
local function mark_goomba_as_dead(obj)
    if obj.parentObj and obj.parentObj ~= obj then
        set_object_respawn_info_bits(obj.parentObj, (obj.oBehParams2ndByte & GOOMBA_BP_TRIPLET_FLAG_MASK) >> 2)
        obj.parentObj.oBehParams = obj.parentObj.oBehParams | (obj.oBehParams2ndByte & GOOMBA_BP_TRIPLET_FLAG_MASK) << 6
    end
end

---@param obj Object
local function goomba_begin_jump(obj)
    cur_obj_play_sound_2(SOUND_OBJ_GOOMBA_ALERT)
    obj.oAction = GOOMBA_ACT_JUMP
    obj.oForwardVel = 0.0
    obj.oVelY = 50.0 / 3.0 * obj.oGoombaScale
    obj.oGoombaJumpCooldown = 10
end

---@param obj Object
local function goomba_act_walk(obj)
    treat_far_home_as_mario(obj, 1000.0, nil, nil)

    -- ! SPEED
    -- Move faster than normal
    obj_forward_vel_approach((obj.oGoombaRelativeSpeed * obj.oGoombaScale) * speed_mutliplier, 0.4 * speed_mutliplier)

    if obj.oGoombaRelativeSpeed > 4/3 then
        cur_obj_play_sound_at_anim_range(2, 17, SOUND_OBJ_GOOMBA_WALK)
    end

    if obj.oGoombaTurningAwayFromWall ~= 0 then
        obj.oGoombaTurningAwayFromWall = obj_resolve_collisions_and_turn(obj.oGoombaTargetYaw, 0x200)
    else
        local player = nearest_player_to_object(obj)
        ---@type number?
        local distanceToPlayer = player and dist_between_objects(obj, player) or 10000
        ---@type number?
        local angleToPlayer = player and obj_angle_to_object(obj, player) or 0
        distanceToPlayer, angleToPlayer = treat_far_home_as_mario(obj, 1000.0, distanceToPlayer, angleToPlayer)

        if distanceToPlayer >= 25000 then
            obj.oGoombaTargetYaw = angleToPlayer
            obj.oGoombaWalkTimer = random_linear_offset(20, 30)
        end

        obj.oGoombaTurningAwayFromWall = obj_bounce_off_walls_edges_objects(get_temp_s32_pointer(obj.oGoombaTargetYaw))
        if obj.oGoombaTurningAwayFromWall == 0 then
            if distanceToPlayer < 500.0 then
                if obj.oGoombaRelativeSpeed <= 2.0 then
                    goomba_begin_jump(obj)
                end

                obj.oGoombaTargetYaw = angleToPlayer
                obj.oGoombaRelativeSpeed = 20.0 * speed_mutliplier
            else
                obj.oGoombaRelativeSpeed = 4/3
                if obj.oGoombaWalkTimer ~= 0 then
                    obj.oGoombaWalkTimer = obj.oGoombaWalkTimer - 1
                else
                    if random_linear_offset(0, 32768) & 3 or obj.oGoombaJumpCooldown > 0 then
                        obj.oGoombaTargetYaw = obj_random_fixed_turn(0x2000)
                        obj.oGoombaWalkTimer = random_linear_offset(100, 100)
                    else
                        goomba_begin_jump(obj)
                        obj.oGoombaTargetYaw = obj_random_fixed_turn(0x6000)
                    end
                end
            end
        end

        cur_obj_rotate_yaw_toward(obj.oGoombaTargetYaw, 0x200)
    end

    if obj.oGoombaJumpCooldown > 0 then
        obj.oGoombaJumpCooldown = obj.oGoombaJumpCooldown - 1
    end
end

---@param obj Object
local function goomba_act_attacked_mario(obj)
    if obj.oGoombaSize == GOOMBA_SIZE_TINY then
        mark_goomba_as_dead(obj)
        obj.oNumLootCoins = 0
        obj_die_if_health_non_positive()
    else
        local player = nearest_player_to_object(obj)
        local angleToPlayer = player and obj_angle_to_object(obj, player) or 0
        goomba_begin_jump(obj)
        obj.oGoombaTargetYaw = angleToPlayer
        obj.oGoombaTurningAwayFromWall = 0
    end
end

---@param obj Object
local function goomba_act_jump(obj)
    obj_resolve_object_collisions(nil)

    if obj.oMoveFlags & OBJ_MOVE_MASK_ON_GROUND ~= 0 then
        obj.oAction = GOOMBA_ACT_WALK
    else
        cur_obj_rotate_yaw_toward(obj.oGoombaTargetYaw, 0x800)
    end
end

---@param obj Object
function bhv_custom_goomba_init(obj)
    obj.oAnimations = gObjectAnimations.goomba_seg8_anims_0801DA4C
    obj.oGoombaSize = obj.oBehParams2ndByte & GOOMBA_BP_SIZE_MASK

    obj.oGoombaScale = sGoombaProperties[obj.oGoombaSize + 1].scale
    obj.oDeathSound = sGoombaProperties[obj.oGoombaSize + 1].deathSound

    obj_set_hitbox(obj, sGoombaHitbox)

    obj.oDrawingDistance = sGoombaProperties[obj.oGoombaSize + 1].drawDistance
    obj.oDamageOrCoinValue = sGoombaProperties[obj.oGoombaSize + 1].damage

    obj.oGravity = -8.0 / 3.0 * obj.oGoombaScale

    network_init_object(obj, true, {
        "oGoombaTargetYaw",
        "oGoombaWalkTimer"
    })
end

---@param obj Object
function bhv_custom_goomba_update(obj)
    local animSpeed = 0

    if obj_update_standard_actions(obj.oGoombaScale) ~= 0 then
        if obj.parentObj and obj.parentObj ~= obj then
            if obj.parentObj.oAction == GOOMBA_TRIPLET_SPAWNER_ACT_UNLOADED then
                obj_mark_for_deletion(obj)
            end
        end

        cur_obj_scale(obj.oGoombaScale)
        obj_update_blinking(get_temp_s32_pointer(obj.oGoombaBlinkTimer), 30, 50, 5)
        cur_obj_update_floor_and_walls()

        animSpeed = obj.oForwardVel / obj.oGoombaScale * 0.4
        if animSpeed < 1.0 then
            animSpeed = 1.0
        end
        cur_obj_init_animation_with_accel_and_sound(0, animSpeed)

        switch (obj.oAction, {
            [GOOMBA_ACT_WALK] = function ()
                goomba_act_walk(obj)
            end,
            [GOOMBA_ACT_ATTACKED_MARIO] = function ()
                goomba_act_attacked_mario(obj)
            end,
            [GOOMBA_ACT_JUMP] = function ()
                goomba_act_jump(obj)
            end,
        })

        if obj_handle_attacks(obj, sGoombaHitbox, GOOMBA_ACT_ATTACKED_MARIO, sGoombaAttackHandlers[(obj.oGoombaSize & 1) + 1]) then
            mark_goomba_as_dead(obj)
        end

        cur_obj_move_standard(-78)
    else
        obj.oAnimState = 1
    end
end